package mhtree.benchmarking;

import cz.muni.fi.disa.similarityoperators.cover.AbstractRepresentation;
import messif.algorithms.AlgorithmMethodException;
import messif.buckets.BucketStorageException;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFloatVectorNeuralNetworkL2;
import messif.objects.util.AbstractObjectList;
import messif.objects.util.RankedAbstractObject;
import messif.objects.util.StreamGenericAbstractObjectIterator;
import messif.operations.Approximate;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.ApproxKNNQueryOperation;
import messif.operations.query.KNNQueryOperation;
import messif.statistics.Statistics;
import mhtree.InsertType;
import mhtree.MHTree;
import mhtree.MergeType;
import mhtree.ObjectToNodeDistance;
import mtree.MTree;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class RunBenchmark {
    public static void main(String[] args) throws IOException, AlgorithmMethodException, InstantiationException, NoSuchMethodException, BucketStorageException {
        if (args.length != 5) {
            throw new IllegalArgumentException("Unexpected number of params");
        }

        Statistics.enableGlobally();
        AbstractRepresentation.PrecomputedDistances.COMPUTATION_THREADS = 16;

        List<LocalAbstractObject> objects = loadDataset(args[0]);
        int leafCapacity = Integer.parseInt(args[1]);
        int nodeDegree = Integer.parseInt(args[2]);
        InsertType insertType = args[3].equals("INCREMENTAL") ? InsertType.INCREMENTAL : InsertType.GREEDY;
        ObjectToNodeDistance objectToNodeDistance;

        switch (args[4]) {
            case "FURTHEST":
                objectToNodeDistance = ObjectToNodeDistance.FURTHEST;
                break;
            case "AVERAGE":
                objectToNodeDistance = ObjectToNodeDistance.AVERAGE;
                break;
            default:
                objectToNodeDistance = ObjectToNodeDistance.NEAREST;
                break;
        }

        percentageToRecallMHTree(new MHTreeConfig(
                        leafCapacity,
                        nodeDegree,
                        insertType,
                        objectToNodeDistance
                ),
                objects,
                new int[]{1, 10, 50, 100}
        );
    }

    private static void percentageToRecallMHTree(MHTreeConfig config, List<LocalAbstractObject> objects, int[] ks) throws BucketStorageException, RuntimeException {
        MHTree mTree = new MHTree.Builder(objects, config.leafCapacity, config.nodeDegree)
                .objectToNodeDistance(config.objectToNodeDistance)
                .mergeType(MergeType.REPRESENTATION_BASED)
                .build();

        mTree.printStatistics();

        System.out.println("leafCapacity,nodeDegree,objectToNodeDistance,k,percentage,recall (min),recall (avg),recall (med),recall (max)");

        int maxK = Arrays.stream(ks).max().getAsInt();

        List<KNNQueryOperation> kNNOperations = objects
                .parallelStream()
                .map(object -> new KNNQueryOperation(object, maxK))
                .collect(Collectors.toList());

        class Pair {
            public List<RankedAbstractObject> kNNObjects;
            public String id;
        }

        Map<String, List<RankedAbstractObject>> kNNResults = kNNOperations
                .parallelStream()
                .map(op -> {
                    mTree.kNN(op);
                    Pair pair = new Pair();
                    pair.id = op.getQueryObject().getLocatorURI();
                    pair.kNNObjects = new ArrayList<>(maxK);
                    for (RankedAbstractObject o : op)
                        pair.kNNObjects.add(o);
                    return pair;
                })
                .collect(Collectors.toMap(entry -> entry.id, entry -> entry.kNNObjects));

        for (int k : ks) {
            List<ApproxKNNQueryOperation> approxOperations = objects
                    .parallelStream()
                    .map(object -> new ApproxKNNQueryOperation(object, k, 0, Approximate.LocalSearchType.PERCENTAGE, LocalAbstractObject.UNKNOWN_DISTANCE))
                    .collect(Collectors.toList());
            approxOperations
                    .parallelStream()
                    .forEach(op -> op.suppData = new SearchState(mTree, op));

            double minimalRecall = 0;
            int percentage = 0;
            int percentageStep = 5;

            while (minimalRecall != 1.0) {
                approxOperations
                        .parallelStream()
                        .filter(op -> ((SearchState) op.suppData).recall != 1d)
                        .forEach(op -> {
                            SearchState searchState = (SearchState) op.suppData;
                            mTree.approxKNN(op);
                            searchState.recall = PerformanceMeasures.measureRecall(op, kNNResults);
                            searchState.approxState.limit += Math.round((float) objects.size() * (float) percentageStep / 100f);
                        });

                Stats recallStats = new Stats(
                        approxOperations
                                .stream()
                                .map(op -> ((SearchState) op.suppData).recall)
                                .collect(Collectors.toList())
                );

                System.out.println(String.join(",",
                        String.valueOf(config.leafCapacity),
                        String.valueOf(config.nodeDegree),
                        String.valueOf(config.objectToNodeDistance),
                        String.valueOf(k),
                        String.valueOf(percentage),
                        String.format("%.2f,%.2f,%.2f,%.2f",
                                recallStats.getMin(),
                                recallStats.getAverage(),
                                recallStats.getMedian(),
                                recallStats.getMax())));

                minimalRecall = recallStats.getMin();
                percentage += percentageStep;
            }
        }
    }

    private static void percentageToRecallMTree(MHTreeConfig config, List<LocalAbstractObject> objects, int[] ks) throws NoSuchMethodException, AlgorithmMethodException, RuntimeException, InstantiationException {
        int numberOfObjects = objects.size();

        MTree mTree = new MTree(config.nodeDegree, config.leafCapacity);

        Collections.shuffle(objects);

        BulkInsertOperation op = new BulkInsertOperation(objects);

        mTree.insert(op);

        mTree.printStatistics();

        System.out.println("leafCapacity,nodeDegree,objectToNodeDistance,k,percentage,recall (min),recall (avg),recall (med),recall (max)");

        double minimalRecall = 0;
        int percentage = 0;
        int percentageStep = 5;

        for (int k : ks) {
            List<Double> recalls = new ArrayList<>(numberOfObjects);
            for (int i = 0; i < numberOfObjects; i++) {
                recalls.add(0.0);
            }

            while (minimalRecall != 1.0) {
                for (int i = 0; i < numberOfObjects; i++) {
                    if (recalls.get(i) != 1.0) {
                        ApproxKNNQueryOperation operation = new ApproxKNNQueryOperation(
                                objects.get(i),
                                k,
                                percentage,
                                Approximate.LocalSearchType.PERCENTAGE,
                                LocalAbstractObject.UNKNOWN_DISTANCE
                        );

                        mTree.executeOperation(operation);

                        recalls.set(i, PerformanceMeasures.measureRecall(operation, mTree));
                    }
                }

                Stats recallStats = new Stats(new ArrayList<>(recalls));

                System.out.println(String.join(",",
                        String.valueOf(config.leafCapacity),
                        String.valueOf(config.nodeDegree),
                        String.valueOf(config.objectToNodeDistance),
                        String.valueOf(k),
                        String.valueOf(percentage),
                        String.format("%.2f,%.2f,%.2f,%.2f",
                                recallStats.getMin(),
                                recallStats.getAverage(),
                                recallStats.getMedian(),
                                recallStats.getMax())));

                minimalRecall = recallStats.getMin();
                percentage += percentageStep;
            }

            minimalRecall = 0;
            percentage = 0;
        }
    }


    private static List<LocalAbstractObject> loadDataset(String path) throws IOException {
        return new AbstractObjectList<>(new StreamGenericAbstractObjectIterator<>(ObjectFloatVectorNeuralNetworkL2.class, path));
    }
}
